import contextlib
import logging
import re
from abc import ABC
from abc import abstractmethod

from pdf2zh_next.config.model import SettingsModel
from pdf2zh_next.translator.base_rate_limiter import BaseRateLimiter
from pdf2zh_next.translator.cache import TranslationCache

logger = logging.getLogger(__name__)


class BaseTranslator(ABC):
    # Due to cache limitations, name should be within 20 characters.
    # cache.py: translate_engine = CharField(max_length=20)
    """translator 的基类，所有的 translator 的实现都需要继承"""

    name = "base"
    lang_map = {}

    def __init__(
        self,
        settings: SettingsModel,
        rate_limiter: BaseRateLimiter,
    ):
        """
        translator class initialization
        :param settings: runtime setting and configuration
        :param rate_limiter: LLM request rate control
        :return: None
        """
        self.ignore_cache = settings.translation.ignore_cache
        lang_in = self.lang_map.get(
            settings.translation.lang_in.lower(), settings.translation.lang_in
        )
        lang_out = self.lang_map.get(
            settings.translation.lang_out.lower(), settings.translation.lang_out
        )
        self.lang_in = lang_in
        self.lang_out = lang_out
        self.rate_limiter = rate_limiter

        self.cache = TranslationCache(
            self.name,
            {
                "lang_in": lang_in,
                "lang_out": lang_out,
            },
        )

        self.translate_call_count = 0
        self.translate_cache_call_count = 0

    def __del__(self):
        with contextlib.suppress(Exception):
            logger.info(
                f"{self.name} translate call count: {self.translate_call_count}"
            )
            logger.info(
                f"{self.name} translate cache call count: {self.translate_cache_call_count}",
            )

    def add_cache_impact_parameters(self, k: str, v):
        """
        Add parameters that affect the translation quality to distinguish the translation effects under different parameters.
        :param k: key
        :param v: value
        """
        self.cache.add_params(k, v)

    def translate(self, text, ignore_cache=False, rate_limit_params: dict = None):
        """
        Translate the text, and the other part should call this method.
        :param text: text to translate
        :return: translated text
        """
        self.translate_call_count += 1
        if not (self.ignore_cache or ignore_cache):
            try:
                cache = self.cache.get(text)
                if cache is not None:
                    self.translate_cache_call_count += 1
                    return cache
            except Exception as e:
                logger.debug(f"try get cache failed, ignore it: {e}")
        self.rate_limiter.wait(rate_limit_params)
        translation = self.do_translate(text)
        if not (self.ignore_cache or ignore_cache):
            self.cache.set(text, translation)
        return translation

    def llm_translate(self, text, ignore_cache=False, rate_limit_params: dict = None):
        """
        Translate the text, and the other part should call this method.
        :param text: text to translate
        :return: translated text
        """
        self.translate_call_count += 1
        if not (self.ignore_cache or ignore_cache):
            try:
                cache = self.cache.get(text)
                if cache is not None:
                    self.translate_cache_call_count += 1
                    return cache
            except Exception as e:
                logger.debug(f"try get cache failed, ignore it: {e}")
        self.rate_limiter.wait(rate_limit_params)
        translation = self.do_llm_translate(text)
        if not (self.ignore_cache or ignore_cache):
            self.cache.set(text, translation)
        return translation

    def do_llm_translate(self, text, rate_limit_params: dict = None):
        """
        Actual translate text, override this method
        :param text: text to translate
        :return: translated text
        """
        raise NotImplementedError

    @abstractmethod
    def do_translate(self, text, rate_limit_params: dict = None):
        """
        Actual translate text, override this method
        :param text: text to translate
        :return: translated text
        """
        logger.critical(
            f"Do not call BaseTranslator.do_translate. "
            f"Translator: {self}. "
            f"Text: {text}. ",
        )
        raise NotImplementedError

    def _remove_cot_content(self, content: str) -> str:
        """Remove text content with the thought chain from the chat response

        :param content: Non-streaming text content
        :return: Text without a thought chain
        """
        return re.sub(r"^<think>.+?</think>", "", content, count=1, flags=re.DOTALL)

    def __str__(self):
        """
        get translator's info
        """
        return f"{self.name} {self.lang_in} {self.lang_out} {self.model}"

    def get_formular_placeholder(self, placeholder_id: int):
        """
        get formular placeholder
        LLM translator use placeholder to skip the formular char
        :param placeholder_id: placeholder id
        :return formated placeholder and regex placeholder
        """
        return "{v" + str(placeholder_id) + "}", f"{{\\s*v\\s*{placeholder_id}\\s*}}"

    def get_rich_text_left_placeholder(self, placeholder_id: int):
        """
        get rich text placeholder
        :param placeholder_id: placeholder id
        :return the start label of rich text and regex start label
        """
        return (
            f"<style id='{placeholder_id}'>",
            f"<\\s*style\\s*id\\s*=\\s*'\\s*{placeholder_id}\\s*'\\s*>",
        )

    def get_rich_text_right_placeholder(self, placeholder_id: int):
        """
        get rich text placeholder
        :return the end label of rich text and regex end label
        """
        return "</style>", r"<\s*\/\s*style\s*>"

    def prompt(self, text):
        """
        concatent the prompt
        :param text: input text
        :return: the whole prompt for LLM translator
        """
        return [
            {
                "role": "user",
                "content": f"You are a professional,authentic machine translation engine.\n\n;; Treat next line as plain text input and translate it into {self.lang_out}, output translation ONLY. If translation is unnecessary (e.g. proper nouns, codes, {'{{1}}, etc. '}), return the original text. NO explanations. NO notes. Input:\n\n{text}",
            },
        ]



import logging

import ollama
from babeldoc.document_il.utils.atomic_integer import AtomicInteger
from pdf2zh_next.config.model import SettingsModel
from pdf2zh_next.translator.base_rate_limiter import BaseRateLimiter
from pdf2zh_next.translator.base_translator import BaseTranslator
from tenacity import before_sleep_log
from tenacity import retry
from tenacity import retry_if_exception_type
from tenacity import stop_after_attempt
from tenacity import wait_exponential

logger = logging.getLogger(__name__)


class OllamaTranslator(BaseTranslator):
    # https://github.com/ollama/ollama
    name = "ollama"

    def __init__(
        self,
        settings: SettingsModel,
        rate_limiter: BaseRateLimiter,
    ):
        super().__init__(settings, rate_limiter)
        self.options = {
            "temperature": 0,
            "num_predict": settings.translate_engine_settings.num_predict,
        }  # 随机采样可能会打断公式标记
        self.client = ollama.Client(
            host=settings.translate_engine_settings.ollama_host,
        )
        self.add_cache_impact_parameters("temperature", self.options["temperature"])
        self.add_cache_impact_parameters("num_predict", self.options["num_predict"])
        self.model = settings.translate_engine_settings.ollama_model
        self.add_cache_impact_parameters("model", self.model)
        self.add_cache_impact_parameters("prompt", self.prompt(""))
        self.token_count = AtomicInteger()
        self.prompt_token_count = AtomicInteger()
        self.completion_token_count = AtomicInteger()

    @retry(
        retry=retry_if_exception_type(ollama.ResponseError),
        stop=stop_after_attempt(100),
        wait=wait_exponential(multiplier=1, min=1, max=15),
        before_sleep=before_sleep_log(logger, logging.WARNING),
    )
    def do_translate(self, text, rate_limit_params: dict = None) -> str:
        if (max_token := len(text) * 5) > self.options["num_predict"]:
            self.options["num_predict"] = max_token
        response = self.client.chat(
            model=self.model,
            options=self.options,
            messages=self.prompt(text),
        )
        self.token_count.inc(response.prompt_eval_count + response.eval_count)
        self.prompt_token_count.inc(response.prompt_eval_count)
        self.completion_token_count.inc(response.eval_count)
        message = response.message.content.strip()
        message = self._remove_cot_content(message)
        return message

    @retry(
        retry=retry_if_exception_type(ollama.ResponseError),
        stop=stop_after_attempt(100),
        wait=wait_exponential(multiplier=1, min=1, max=15),
        before_sleep=before_sleep_log(logger, logging.WARNING),
    )
    def do_llm_translate(self, text, rate_limit_params: dict = None):
        if text is None:
            return None

        if (max_token := len(text) * 5) > self.options["num_predict"]:
            self.options["num_predict"] = max_token

        response = self.client.chat(
            model=self.model,
            options=self.options,
            messages=[
                {
                    "role": "user",
                    "content": text,
                },
            ],
        )
        self.token_count.inc(response.prompt_eval_count + response.eval_count)
        self.prompt_token_count.inc(response.prompt_eval_count)
        self.completion_token_count.inc(response.eval_count)
        message = response.message.content.strip()
        message = self._remove_cot_content(message)
        return message



import asyncio
import cgi
import logging
import shutil
import typing
import uuid
from pathlib import Path
from string import Template

import gradio as gr
import requests
from babeldoc import __version__ as babeldoc_version
from gradio_pdf import PDF

from pdf2zh_next import __version__
from pdf2zh_next.config import ConfigManager
from pdf2zh_next.config.cli_env_model import CLIEnvSettingsModel
from pdf2zh_next.config.model import SettingsModel
from pdf2zh_next.config.translate_engine_model import GUI_PASSWORD_FIELDS
from pdf2zh_next.config.translate_engine_model import GUI_SENSITIVE_FIELDS
from pdf2zh_next.config.translate_engine_model import TRANSLATION_ENGINE_METADATA
from pdf2zh_next.config.translate_engine_model import TRANSLATION_ENGINE_METADATA_MAP
from pdf2zh_next.high_level import TranslationError
from pdf2zh_next.high_level import do_translate_async_stream

logger = logging.getLogger(__name__)
__gui_service_arg_names = []
# The following variables associate strings with specific languages
lang_map = {
    "English": "en",
    "Simplified Chinese": "zh-CN",
    "Traditional Chinese - Hong Kong": "zh-HK",
    "Traditional Chinese - Taiwan": "zh-TW",
    "Japanese": "ja",
    "Korean": "ko",
    "Polish": "pl",
    "Russian": "ru",
    "Spanish": "es",
    "Portuguese": "pt",
    "Brazilian Portuguese": "pt-BR",
    "French": "fr",
    "Malay": "ms",
    "Indonesian": "id",
    "Turkmen": "tk",
    "Filipino (Tagalog)": "tl",
    "Vietnamese": "vi",
    "Kazakh (Latin)": "kk",
    "German": "de",
    "Dutch": "nl",
    "Irish": "ga",
    "Italian": "it",
    "Greek": "el",
    "Swedish": "sv",
    "Danish": "da",
    "Norwegian": "no",
    "Icelandic": "is",
    "Finnish": "fi",
    "Ukrainian": "uk",
    "Czech": "cs",
    "Romanian": "ro",  # Covers Romanian, Moldovan, Moldovan (Cyrillic)
    "Hungarian": "hu",
    "Slovak": "sk",
    "Croatian": "hr",  # Also listed later, keep first
    "Estonian": "et",
    "Latvian": "lv",
    "Lithuanian": "lt",
    "Belarusian": "be",
    "Macedonian": "mk",
    "Albanian": "sq",
    "Serbian (Cyrillic)": "sr",  # Covers Serbian (Latin) too
    "Slovenian": "sl",
    "Catalan": "ca",
    "Bulgarian": "bg",
    "Maltese": "mt",
    "Swahili": "sw",
    "Amharic": "am",
    "Oromo": "om",
    "Tigrinya": "ti",
    "Haitian Creole": "ht",
    "Latin": "la",
    "Lao": "lo",
    "Malayalam": "ml",
    "Gujarati": "gu",
    "Thai": "th",
    "Burmese": "my",
    "Tamil": "ta",
    "Telugu": "te",
    "Oriya": "or",  # Also listed later, keep first
    "Armenian": "hy",
    "Mongolian (Cyrillic)": "mn",
    "Georgian": "ka",
    "Khmer": "km",
    "Bosnian": "bs",
    "Luxembourgish": "lb",
    "Romansh": "rm",
    "Turkish": "tr",
    "Sinhala": "si",
    "Uzbek": "uz",
    "Kyrgyz": "ky",  # Listed as Kirghiz later, keep this one
    "Tajik": "tg",
    "Abkhazian": "ab",
    "Afar": "aa",
    "Afrikaans": "af",
    "Akan": "ak",
    "Aragonese": "an",
    "Avaric": "av",
    "Ewe": "ee",
    "Aymara": "ay",
    "Ojibwa": "oj",
    "Occitan": "oc",
    "Ossetian": "os",
    "Pali": "pi",
    "Bashkir": "ba",
    "Basque": "eu",
    "Breton": "br",
    "Chamorro": "ch",
    "Chechen": "ce",
    "Chuvash": "cv",
    "Tswana": "tn",
    "Ndebele, South": "nr",
    "Ndonga": "ng",
    "Faroese": "fo",
    "Fijian": "fj",
    "Frisian, Western": "fy",
    "Ganda": "lg",
    "Kongo": "kg",
    "Kalaallisut": "kl",
    "Church Slavic": "cu",
    "Guarani": "gn",
    "Interlingua": "ia",
    "Herero": "hz",
    "Kikuyu": "ki",
    "Rundi": "rn",
    "Kinyarwanda": "rw",
    "Galician": "gl",
    "Kanuri": "kr",
    "Cornish": "kw",
    "Komi": "kv",
    "Xhosa": "xh",
    "Corsican": "co",
    "Cree": "cr",
    "Quechua": "qu",
    "Kurdish (Latin)": "ku",
    "Kuanyama": "kj",
    "Limburgan": "li",
    "Lingala": "ln",
    "Manx": "gv",
    "Malagasy": "mg",
    "Marshallese": "mh",
    "Maori": "mi",
    "Navajo": "nv",
    "Nauru": "na",
    "Nyanja": "ny",
    "Norwegian Nynorsk": "nn",
    "Sardinian": "sc",
    "Northern Sami": "se",
    "Samoan": "sm",
    "Sango": "sg",
    "Shona": "sn",
    "Esperanto": "eo",
    "Scottish Gaelic": "gd",
    "Somali": "so",
    "Southern Sotho": "st",
    "Tatar": "tt",
    "Tahitian": "ty",
    "Tongan": "to",
    "Twi": "tw",
    "Walloon": "wa",
    "Welsh": "cy",
    "Venda": "ve",
    "Volapük": "vo",
    "Interlingue": "ie",
    "Hiri Motu": "ho",
    "Igbo": "ig",
    "Ido": "io",
    "Inuktitut": "iu",
    "Inupiaq": "ik",
    "Sichuan Yi": "ii",
    "Yoruba": "yo",
    "Zhuang": "za",
    "Tsonga": "ts",
    "Zulu": "zu",
}

rev_lang_map = {v: k for k, v in lang_map.items()}

# The following variable associate strings with page ranges
page_map = {
    "All": None,
    "First": [0],
    "First 5 pages": list(range(0, 5)),
    "Range": None,  # User-defined range
}

# Load configuration
config_manager = ConfigManager()
try:
    # Load configuration from files and environment variables
    settings = config_manager.initialize_cli_config()
    # Check if sensitive inputs should be disabled in GUI
    disable_sensitive_input = settings.gui_settings.disable_gui_sensitive_input
except Exception as e:
    logger.warning(f"Could not load initial config: {e}")
    settings = CLIEnvSettingsModel()
    disable_sensitive_input = False

# Define default values
default_lang_from = rev_lang_map.get(settings.translation.lang_in, "English")

default_lang_to = settings.translation.lang_out
for display_name, code in lang_map.items():
    if code == default_lang_to:
        default_lang_to = display_name
        break
else:
    default_lang_to = "Simplified Chinese"  # Fallback

# Available translation services
# This will eventually be dynamically determined based on available translators
available_services = [x.translate_engine_type for x in TRANSLATION_ENGINE_METADATA]

if settings.gui_settings.enabled_services:
    enabled_services = {
        x.lower() for x in settings.gui_settings.enabled_services.split(",")
    }
    available_services = [
        x for x in available_services if x.lower() in enabled_services
    ]

assert available_services, "No translation service is enabled"


disable_gui_sensitive_input = settings.gui_settings.disable_gui_sensitive_input


def download_with_limit(url: str, save_path: str, size_limit: int = None) -> str:
    """
    This function downloads a file from a URL and saves it to a specified path.

    Inputs:
        - url: The URL to download the file from
        - save_path: The path to save the file to
        - size_limit: The maximum size of the file to download

    Returns:
        - The path of the downloaded file
    """
    chunk_size = 1024
    total_size = 0
    with requests.get(url, stream=True, timeout=10) as response:
        response.raise_for_status()
        content = response.headers.get("Content-Disposition")
        try:  # filename from header
            _, params = cgi.parse_header(content)
            filename = params["filename"]
        except Exception:  # filename from url
            filename = Path(url).name
        filename = Path(filename).stem + ".pdf"
        save_path = Path(save_path)
        file_path = save_path / filename
        with file_path.open("wb") as file:
            for chunk in response.iter_content(chunk_size=chunk_size):
                total_size += len(chunk)
                if size_limit and total_size > size_limit:
                    raise gr.Error("Exceeds file size limit")
                file.write(chunk)
    return file_path


def _prepare_input_file(
    file_type: str, file_input: str, link_input: str, output_dir: Path
) -> Path:
    """
    This function prepares the input file for translation.

    Inputs:
        - file_type: The type of file to translate (File or Link)
        - file_input: The path to the file to translate
        - link_input: The link to the file to translate
        - output_dir: The directory to save the file to

    Returns:
        - The path of the input file
    """
    if file_type == "File":
        if not file_input:
            raise gr.Error("No file input provided")
        file_path = shutil.copy(file_input, output_dir)
    else:
        if not link_input:
            raise gr.Error("No link input provided")
        try:
            file_path = download_with_limit(link_input, output_dir)
        except Exception as e:
            raise gr.Error(f"Error downloading file: {e}") from e

    return Path(file_path)


def _build_translate_settings(
    base_settings: CLIEnvSettingsModel,
    file_path: Path,
    output_dir: Path,
    ui_inputs: dict,
) -> SettingsModel:
    """
    This function builds translation settings from UI inputs.

    Inputs:
        - base_settings: The base settings model to build upon
        - file_path: The path to the input file
        - output_dir: The output directory
        - ui_inputs: A dictionary of UI inputs

    Returns:
        - A configured SettingsModel instance
    """
    # Clone base settings to avoid modifying the original
    translate_settings = base_settings.clone()
    original_output = translate_settings.translation.output
    original_pages = translate_settings.pdf.pages
    original_gui_settings = config_manager.config_cli_settings.gui_settings

    # Extract UI values
    service = ui_inputs.get("service")
    lang_from = ui_inputs.get("lang_from")
    lang_to = ui_inputs.get("lang_to")
    page_range = ui_inputs.get("page_range")
    page_input = ui_inputs.get("page_input")
    prompt = ui_inputs.get("prompt")
    threads = ui_inputs.get("threads")
    ignore_cache = ui_inputs.get("ignore_cache")

    # PDF Output Options
    no_mono = ui_inputs.get("no_mono")
    no_dual = ui_inputs.get("no_dual")
    dual_translate_first = ui_inputs.get("dual_translate_first")
    use_alternating_pages_dual = ui_inputs.get("use_alternating_pages_dual")
    watermark_output_mode = ui_inputs.get("watermark_output_mode")

    # Advanced Translation Options
    min_text_length = ui_inputs.get("min_text_length")
    rpc_doclayout = ui_inputs.get("rpc_doclayout")

    # Advanced PDF Options
    skip_clean = ui_inputs.get("skip_clean")
    disable_rich_text_translate = ui_inputs.get("disable_rich_text_translate")
    enhance_compatibility = ui_inputs.get("enhance_compatibility")
    split_short_lines = ui_inputs.get("split_short_lines")
    short_line_split_factor = ui_inputs.get("short_line_split_factor")
    translate_table_text = ui_inputs.get("translate_table_text")
    skip_scanned_detection = ui_inputs.get("skip_scanned_detection")
    ocr_workaround = ui_inputs.get("ocr_workaround")
    max_pages_per_part = ui_inputs.get("max_pages_per_part")
    formular_font_pattern = ui_inputs.get("formular_font_pattern")
    formular_char_pattern = ui_inputs.get("formular_char_pattern")

    # New input for custom_system_prompt
    custom_system_prompt_input = ui_inputs.get("custom_system_prompt_input")

    # Map UI language selections to language codes
    source_lang = lang_map.get(lang_from, "auto")
    target_lang = lang_map.get(lang_to, "zh")

    # Set up page selection
    if page_range == "Range" and page_input:
        pages = page_input  # The backend parser handles the format
    else:
        # Use predefined ranges from page_map
        selected_pages = page_map[page_range]
        if selected_pages is None:
            pages = None  # All pages
        else:
            # Convert page indices to comma-separated string
            pages = ",".join(
                str(p + 1) for p in selected_pages
            )  # +1 because UI is 1-indexed

    # Update settings with UI values
    translate_settings.basic.input_files = {str(file_path)}
    translate_settings.report_interval = 0.2
    translate_settings.translation.lang_in = source_lang
    translate_settings.translation.lang_out = target_lang
    translate_settings.translation.output = str(output_dir)
    translate_settings.translation.qps = int(threads)
    translate_settings.translation.ignore_cache = ignore_cache

    # Update Translation Settings
    if min_text_length is not None:
        translate_settings.translation.min_text_length = int(min_text_length)
    if rpc_doclayout:
        translate_settings.translation.rpc_doclayout = rpc_doclayout

    # Update PDF Settings
    translate_settings.pdf.pages = pages
    translate_settings.pdf.no_mono = no_mono
    translate_settings.pdf.no_dual = no_dual
    translate_settings.pdf.dual_translate_first = dual_translate_first
    translate_settings.pdf.use_alternating_pages_dual = use_alternating_pages_dual

    # Map watermark mode from UI to enum
    if watermark_output_mode == "Watermarked":
        from pdf2zh_next.config.model import WatermarkOutputMode

        translate_settings.pdf.watermark_output_mode = WatermarkOutputMode.Watermarked
    elif watermark_output_mode == "No Watermark":
        from pdf2zh_next.config.model import WatermarkOutputMode

        translate_settings.pdf.watermark_output_mode = WatermarkOutputMode.NoWatermark

    # Update Advanced PDF Settings
    translate_settings.pdf.skip_clean = skip_clean
    translate_settings.pdf.disable_rich_text_translate = disable_rich_text_translate
    translate_settings.pdf.enhance_compatibility = enhance_compatibility
    translate_settings.pdf.split_short_lines = split_short_lines
    translate_settings.pdf.ocr_workaround = ocr_workaround
    if short_line_split_factor is not None:
        translate_settings.pdf.short_line_split_factor = float(short_line_split_factor)

    translate_settings.pdf.translate_table_text = translate_table_text
    translate_settings.pdf.skip_scanned_detection = skip_scanned_detection

    if max_pages_per_part is not None and max_pages_per_part > 0:
        translate_settings.pdf.max_pages_per_part = int(max_pages_per_part)

    if formular_font_pattern:
        translate_settings.pdf.formular_font_pattern = formular_font_pattern

    if formular_char_pattern:
        translate_settings.pdf.formular_char_pattern = formular_char_pattern

    assert service in TRANSLATION_ENGINE_METADATA_MAP, "UNKNOW TRANSLATION ENGINE!"

    for metadata in TRANSLATION_ENGINE_METADATA:
        cli_flag = metadata.cli_flag_name
        setattr(translate_settings, cli_flag, False)

    metadata = TRANSLATION_ENGINE_METADATA_MAP[service]
    cli_flag = metadata.cli_flag_name
    setattr(translate_settings, cli_flag, True)
    if metadata.cli_detail_field_name:
        detail_setting = getattr(translate_settings, metadata.cli_detail_field_name)
        if metadata.setting_model_type:
            for field_name in metadata.setting_model_type.model_fields:
                if field_name == "translate_engine_type":
                    continue
                if disable_gui_sensitive_input:
                    if field_name in GUI_PASSWORD_FIELDS:
                        continue
                    if field_name in GUI_SENSITIVE_FIELDS:
                        continue
                value = ui_inputs.get(field_name)
                type_hint = detail_setting.model_fields[field_name].annotation
                original_type = typing.get_origin(type_hint)
                type_args = typing.get_args(type_hint)
                if type_hint is str or str in type_args:
                    pass
                elif type_hint is int or int in type_args:
                    value = int(value)
                elif type_hint is bool or bool in type_args:
                    value = bool(value)
                else:
                    raise Exception(
                        f"Unsupported type {type_hint} for field {field_name} in gui translation engine settings"
                    )
                setattr(detail_setting, field_name, value)

    # Add custom prompt if provided
    if prompt:
        # This might need adjustment based on how prompt is handled in the new system
        translate_settings.custom_prompt = Template(prompt)

    # Add custom system prompt if provided
    custom_system_prompt_value = ui_inputs.get("custom_system_prompt_input")
    if custom_system_prompt_value:
        translate_settings.translation.custom_system_prompt = custom_system_prompt_value
    else:
        translate_settings.translation.custom_system_prompt = None

    # Validate settings before proceeding
    try:
        translate_settings.validate_settings()
        settings = translate_settings.to_settings_model()
        translate_settings.translation.output = original_output
        translate_settings.pdf.pages = original_pages
        translate_settings.gui_settings = original_gui_settings
        translate_settings.basic.gui = False
        if not settings.gui_settings.disable_config_auto_save:
            config_manager.write_user_default_config_file(settings=translate_settings)
        settings.validate_settings()
        return settings
    except ValueError as e:
        raise gr.Error(f"Invalid settings: {e}") from e


async def _run_translation_task(
    settings: SettingsModel, file_path: Path, state: dict, progress: gr.Progress
) -> tuple[Path | None, Path | None]:
    """
    This function runs the translation task and handles progress updates.

    Inputs:
        - settings: The translation settings
        - file_path: The path to the input file
        - state: The state dictionary for tracking the task
        - progress: The Gradio progress bar

    Returns:
        - A tuple of (mono_pdf_path, dual_pdf_path)
    """
    mono_path = None
    dual_path = None

    try:
        settings.basic.input_files = set()
        async for event in do_translate_async_stream(settings, file_path):
            if event["type"] in (
                "progress_start",
                "progress_update",
                "progress_end",
            ):
                # Update progress bar
                desc = event["stage"]
                progress_value = event["overall_progress"] / 100.0
                part_index = event["part_index"]
                total_parts = event["total_parts"]
                stage_current = event["stage_current"]
                stage_total = event["stage_total"]
                desc = f"{desc} ({part_index}/{total_parts}, {stage_current}/{stage_total})"
                logger.info(f"Progress: {progress_value}, {desc}")
                progress(progress_value, desc=desc)
            elif event["type"] == "finish":
                # Extract result paths
                result = event["translate_result"]
                mono_path = result.mono_pdf_path
                dual_path = result.dual_pdf_path
                progress(1.0, desc="Translation complete!")
                break
            elif event["type"] == "error":
                # Handle error event
                error_msg = event.get("error", "Unknown error")
                error_details = event.get("details", "")
                # error_str = f"{error_msg}" + (
                #     f": {error_details}" if error_details else ""
                # )
                raise gr.Error(f"Translation error: {error_msg}")
    except asyncio.CancelledError:
        # Handle task cancellation - let translate_file handle the UI updates
        logger.info(
            f"Translation for session {state.get('session_id', 'unknown')} was cancelled"
        )
        raise  # Re-raise for the calling function to handle
    except TranslationError as e:
        # Handle structured translation errors
        logger.error(f"Translation error: {e}")
        raise gr.Error(f"Translation error: {e.message}") from e
    except gr.Error as e:
        # Handle Gradio errors
        logger.error(f"Gradio error: {e}")
        raise
    except Exception as e:
        # Handle other exceptions
        logger.error(f"Error in _run_translation_task: {e}", exc_info=True)
        raise gr.Error(f"Translation failed: {e}") from e

    return mono_path, dual_path


async def stop_translate_file(state: dict) -> None:
    """
    This function stops the translation process.

    Inputs:
        - state: The state of the translation process

    Returns:- None
    """
    if "current_task" not in state or state["current_task"] is None:
        return

    logger.info(
        f"Stopping translation for session {state.get('session_id', 'unknown')}"
    )
    # Cancel the task
    try:
        state["current_task"].cancel()
        # Wait briefly for cancellation to take effect
        await asyncio.sleep(0.1)
    except Exception as e:
        logger.error(f"Error stopping translation: {e}")
    finally:
        state["current_task"] = None


async def translate_file(
    file_type,
    file_input,
    link_input,
    service,
    lang_from,
    lang_to,
    page_range,
    page_input,
    # PDF Output Options
    no_mono,
    no_dual,
    dual_translate_first,
    use_alternating_pages_dual,
    watermark_output_mode,
    # Advanced Options
    prompt,
    threads,
    min_text_length,
    rpc_doclayout,
    # New input for custom_system_prompt
    custom_system_prompt_input,
    skip_clean,
    disable_rich_text_translate,
    enhance_compatibility,
    split_short_lines,
    short_line_split_factor,
    translate_table_text,
    skip_scanned_detection,
    max_pages_per_part,
    formular_font_pattern,
    formular_char_pattern,
    ignore_cache,
    state,
    ocr_workaround,
    *translation_engine_arg_inputs,
    progress=None,
):
    """
    This function translates a PDF file from one language to another using the new architecture.

    Inputs:
        - file_type: The type of file to translate
        - file_input: The file to translate
        - link_input: The link to the file to translate
        - service: The translation service to use
        - lang_from: The language to translate from
        - lang_to: The language to translate to
        - page_range: The range of pages to translate
        - page_input: The input for the page range
        - prompt: The custom prompt for the llm
        - threads: The number of threads to use
        - skip_clean: Whether to skip subsetting fonts
        - ignore_cache: Whether to ignore the translation cache
        - state: The state of the translation process
        - translation_engine_arg_inputs: The translator engine args
        - progress: The progress bar

    Returns:
        - The translated mono PDF file
        - The preview PDF file
        - The translated dual PDF file
        - The visibility state of the mono PDF output
        - The visibility state of the dual PDF output
        - The visibility state of the output title
    """
    # Setup progress tracking
    if progress is None:
        progress = gr.Progress()

    # Initialize session and output directory
    session_id = str(uuid.uuid4())
    state["session_id"] = session_id

    # Track progress
    progress(0, desc="Starting translation...")

    # Prepare output directory
    output_dir = Path("pdf2zh_files") / session_id
    output_dir.mkdir(parents=True, exist_ok=True)

    # Collection of UI inputs for config building
    ui_inputs = {
        "service": service,
        "lang_from": lang_from,
        "lang_to": lang_to,
        "page_range": page_range,
        "page_input": page_input,
        # PDF Output Options
        "no_mono": no_mono,
        "no_dual": no_dual,
        "dual_translate_first": dual_translate_first,
        "use_alternating_pages_dual": use_alternating_pages_dual,
        "watermark_output_mode": watermark_output_mode,
        # Advanced Options
        "prompt": prompt,
        "threads": threads,
        "min_text_length": min_text_length,
        "rpc_doclayout": rpc_doclayout,
        "custom_system_prompt_input": custom_system_prompt_input,
        "skip_clean": skip_clean,
        "disable_rich_text_translate": disable_rich_text_translate,
        "enhance_compatibility": enhance_compatibility,
        "split_short_lines": split_short_lines,
        "short_line_split_factor": short_line_split_factor,
        "translate_table_text": translate_table_text,
        "skip_scanned_detection": skip_scanned_detection,
        "max_pages_per_part": max_pages_per_part,
        "formular_font_pattern": formular_font_pattern,
        "formular_char_pattern": formular_char_pattern,
        "ignore_cache": ignore_cache,
        "ocr_workaround": ocr_workaround,
    }
    for arg_name, arg_input in zip(
        __gui_service_arg_names, translation_engine_arg_inputs, strict=False
    ):
        ui_inputs[arg_name] = arg_input
    try:
        # Step 1: Prepare input file
        file_path = _prepare_input_file(file_type, file_input, link_input, output_dir)

        # Step 2: Build translation settings
        translate_settings = _build_translate_settings(
            settings.clone(), file_path, output_dir, ui_inputs
        )

        # Step 3: Create and run the translation task
        task = asyncio.create_task(
            _run_translation_task(translate_settings, file_path, state, progress)
        )
        state["current_task"] = task

        # Wait for the translation to complete
        mono_path, dual_path = await task
        if not mono_path.exists():
            mono_path = None
        else:
            mono_path = mono_path.as_posix()
        if not dual_path or not dual_path.exists():
            dual_path = None
        else:
            dual_path = dual_path.as_posix()
        # Build success UI updates
        return (
            str(mono_path) if mono_path else None,  # Output mono file
            str(mono_path) if mono_path else dual_path,  # Preview
            str(dual_path) if dual_path else None,  # Output dual file
            gr.update(visible=bool(mono_path)),  # Show mono download if available
            gr.update(visible=bool(dual_path)),  # Show dual download if available
            gr.update(
                visible=bool(mono_path or dual_path)
            ),  # Show output title if any output
        )
    except asyncio.CancelledError:
        gr.Info("Translation cancelled")
        # Return None for all outputs if cancelled
        return (
            None,
            None,
            None,
            gr.update(visible=False),
            gr.update(visible=False),
            gr.update(visible=False),
        )
    except gr.Error:
        # Re-raise Gradio errors without modification
        raise
    except Exception as e:
        # Catch any other errors and wrap in gr.Error
        logger.exception(f"Error in translate_file: {e}")
        raise gr.Error(f"Translation failed: {e}") from e
    finally:
        # Clear task reference
        state["current_task"] = None


# Custom theme definition
custom_blue = gr.themes.Color(
    c50="#E8F3FF",
    c100="#BEDAFF",
    c200="#94BFFF",
    c300="#6AA1FF",
    c400="#4080FF",
    c500="#165DFF",  # Primary color
    c600="#0E42D2",
    c700="#0A2BA6",
    c800="#061D79",
    c900="#03114D",
    c950="#020B33",
)

custom_css = """
    .secondary-text {color: #999 !important;}
    footer {visibility: hidden}
    .env-warning {color: #dd5500 !important;}
    .env-success {color: #559900 !important;}

    /* Add dashed border to input-file class */
    .input-file {
        border: 1.2px dashed #165DFF !important;
        border-radius: 6px !important;
    }

    .progress-bar-wrap {
        border-radius: 8px !important;
    }

    .progress-bar {
        border-radius: 8px !important;
    }

    .pdf-canvas canvas {
        width: 100%;
    }
    """

tech_details_string = f"""
                    <summary>Technical details</summary>
                    - GitHub: <a href="https://github.com/PDFMathTranslate/PDFMathTranslate-next">PDFMathTranslate/PDFMathTranslate-next</a><br>
                    - BabelDOC: <a href="https://github.com/funstory-ai/BabelDOC">funstory-ai/BabelDOC</a><br>
                    - GUI by: <a href="https://github.com/reycn">Rongxin</a> & <a href="https://github.com/hellofinch">hellofinch</a> & <a href="https://github.com/awwaawwa">awwaawwa</a><br>
                    - pdf2zh Version: {__version__} <br>
                    - BabelDOC Version: {babeldoc_version}
                """

# The following code creates the GUI
with gr.Blocks(
    title="PDFMathTranslate - PDF Translation with preserved formats",
    theme=gr.themes.Default(
        primary_hue=custom_blue, spacing_size="md", radius_size="lg"
    ),
    css=custom_css,
) as demo:
    gr.Markdown(
        "# [PDFMathTranslate @ GitHub](https://github.com/PDFMathTranslate/PDFMathTranslate-next)"
    )

    translation_engine_arg_inputs = []
    detail_text_inputs = []
    detail_text_input_index_map = {}
    with gr.Row():
        with gr.Column(scale=1):
            gr.Markdown("## File")
            file_type = gr.Radio(
                choices=["File", "Link"],
                label="Type",
                value="File",
            )
            file_input = gr.File(
                label="File",
                file_count="single",
                file_types=[".pdf"],
                type="filepath",
                elem_classes=["input-file"],
            )
            link_input = gr.Textbox(
                label="Link",
                visible=False,
                interactive=True,
            )

            gr.Markdown("## Translation Options")
            detail_index = 0
            with gr.Group() as translation_engine_settings:
                service = gr.Dropdown(
                    label="Service",
                    choices=available_services,
                    value=available_services[0],
                )
                __gui_service_arg_names = []
                for service_name in available_services:
                    metadata = TRANSLATION_ENGINE_METADATA_MAP[service_name]
                    if not metadata.cli_detail_field_name:
                        # no detail field, no need to show
                        continue
                    detail_settings = getattr(settings, metadata.cli_detail_field_name)
                    visible = service.value == metadata.translate_engine_type
                    # OpenAI specific settings (initially visible if OpenAI is default)
                    with gr.Group(visible=True) as service_detail:
                        detail_text_input_index_map[metadata.translate_engine_type] = []

                        for (
                            field_name,
                            field,
                        ) in metadata.setting_model_type.model_fields.items():
                            if disable_gui_sensitive_input:
                                if field_name in GUI_SENSITIVE_FIELDS:
                                    continue
                                if field_name in GUI_PASSWORD_FIELDS:
                                    continue
                            if field.default_factory:
                                continue

                            if field_name == "translate_engine_type":
                                continue
                            type_hint = field.annotation
                            original_type = typing.get_origin(type_hint)
                            type_args = typing.get_args(type_hint)
                            value = getattr(detail_settings, field_name)
                            if (
                                type_hint is str
                                or str in type_args
                                or type_hint is int
                                or int in type_args
                            ):
                                if field_name in GUI_PASSWORD_FIELDS:
                                    field_input = gr.Textbox(
                                        label=field.description,
                                        value=value,
                                        interactive=True,
                                        type="password",
                                        visible=visible,
                                    )
                                else:
                                    field_input = gr.Textbox(
                                        label=field.description,
                                        value=value,
                                        interactive=True,
                                        visible=visible,
                                    )
                            elif type_hint is bool or bool in type_args:
                                field_input = gr.Checkbox(
                                    label=field.description,
                                    value=value,
                                    interactive=True,
                                    visible=visible,
                                )
                            else:
                                raise Exception(
                                    f"Unsupported type {type_hint} for field {field_name} in gui translation engine settings"
                                )
                            detail_text_input_index_map[
                                metadata.translate_engine_type
                            ].append(detail_index)
                            detail_index += 1
                            detail_text_inputs.append(field_input)
                            __gui_service_arg_names.append(field_name)
                            translation_engine_arg_inputs.append(field_input)

            with gr.Row():
                lang_from = gr.Dropdown(
                    label="Translate from",
                    choices=list(lang_map.keys()),
                    value=default_lang_from,
                )
                lang_to = gr.Dropdown(
                    label="Translate to",
                    choices=list(lang_map.keys()),
                    value=default_lang_to,
                )

            page_range = gr.Radio(
                choices=list(page_map.keys()),
                label="Pages",
                value=list(page_map.keys())[0],
            )

            page_input = gr.Textbox(
                label="Page range (e.g., 1,3,5-10,-5)",
                visible=False,
                interactive=True,
                placeholder="e.g., 1,3,5-10",
            )

            # PDF Output Options
            gr.Markdown("## PDF Output Options")
            with gr.Row():
                no_mono = gr.Checkbox(
                    label="Disable monolingual output",
                    value=settings.pdf.no_mono,
                    interactive=True,
                )
                no_dual = gr.Checkbox(
                    label="Disable bilingual output",
                    value=settings.pdf.no_dual,
                    interactive=True,
                )

            with gr.Row():
                dual_translate_first = gr.Checkbox(
                    label="Put translated pages first in dual mode",
                    value=settings.pdf.dual_translate_first,
                    interactive=True,
                )
                use_alternating_pages_dual = gr.Checkbox(
                    label="Use alternating pages for dual PDF",
                    value=settings.pdf.use_alternating_pages_dual,
                    interactive=True,
                )

            watermark_output_mode = gr.Radio(
                choices=["Watermarked", "No Watermark"],
                label="Watermark mode",
                value="Watermarked"
                if settings.pdf.watermark_output_mode.value == "watermarked"
                else "No Watermark",
            )

            # Additional translation options
            with gr.Accordion("Advanced Options", open=False):
                prompt = gr.Textbox(
                    label="Custom prompt for translation",
                    value="",
                    visible=False,
                    interactive=True,
                    placeholder="Custom prompt for the translator",
                )

                threads = gr.Number(
                    label="RPS (Requests Per Second)",
                    value=settings.translation.qps or 4,
                    precision=0,
                    minimum=1,
                    interactive=True,
                )

                # New Textbox for custom_system_prompt
                custom_system_prompt_input = gr.Textbox(
                    label="Custom System Prompt",
                    value=settings.translation.custom_system_prompt or "",
                    interactive=True,
                    placeholder="e.g. /no_think You are a professional, authentic machine translation engine.",
                )

                min_text_length = gr.Number(
                    label="Minimum text length to translate",
                    value=settings.translation.min_text_length,
                    precision=0,
                    minimum=0,
                    interactive=True,
                )

                rpc_doclayout = gr.Textbox(
                    label="RPC service for document layout analysis (optional)",
                    value=settings.translation.rpc_doclayout or "",
                    visible=False,
                    interactive=True,
                    placeholder="http://host:port",
                )

                # PDF options section
                gr.Markdown("### PDF Options")

                skip_clean = gr.Checkbox(
                    label="Skip clean (maybe improve compatibility)",
                    value=settings.pdf.skip_clean,
                    interactive=True,
                )

                disable_rich_text_translate = gr.Checkbox(
                    label="Disable rich text translation (maybe improve compatibility)",
                    value=settings.pdf.disable_rich_text_translate,
                    interactive=True,
                )

                enhance_compatibility = gr.Checkbox(
                    label="Enhance compatibility (auto-enables skip_clean and disable_rich_text)",
                    value=settings.pdf.enhance_compatibility,
                    interactive=True,
                )

                split_short_lines = gr.Checkbox(
                    label="Force split short lines into different paragraphs",
                    value=settings.pdf.split_short_lines,
                    interactive=True,
                )

                short_line_split_factor = gr.Slider(
                    label="Split threshold factor for short lines",
                    value=settings.pdf.short_line_split_factor,
                    minimum=0.1,
                    maximum=1.0,
                    step=0.1,
                    interactive=True,
                    visible=settings.pdf.split_short_lines,
                )

                translate_table_text = gr.Checkbox(
                    label="Translate table text (experimental)",
                    value=settings.pdf.translate_table_text,
                    interactive=True,
                )

                skip_scanned_detection = gr.Checkbox(
                    label="Skip scanned detection",
                    value=settings.pdf.skip_scanned_detection,
                    interactive=True,
                )

                ocr_workaround = gr.Checkbox(
                    label="OCR workaround (experimental, will auto enable Skip scanned detection in backend)",
                    value=settings.pdf.ocr_workaround,
                    interactive=True,
                )

                max_pages_per_part = gr.Number(
                    label="Maximum pages per part (for auto-split translation, 0 means no limit)",
                    value=settings.pdf.max_pages_per_part,
                    precision=0,
                    minimum=0,
                    interactive=True,
                )

                formular_font_pattern = gr.Textbox(
                    label="Font pattern to identify formula text (regex, not recommended to change)",
                    value=settings.pdf.formular_font_pattern or "",
                    interactive=True,
                    placeholder="e.g., CMMI|CMR",
                )

                formular_char_pattern = gr.Textbox(
                    label="Character pattern to identify formula text (regex, not recommended to change)",
                    value=settings.pdf.formular_char_pattern or "",
                    interactive=True,
                    placeholder="e.g., [∫∬∭∮∯∰∇∆]",
                )

                ignore_cache = gr.Checkbox(
                    label="Ignore cache",
                    value=settings.translation.ignore_cache,
                    interactive=True,
                )

            output_title = gr.Markdown("## Translated", visible=False)
            output_file_mono = gr.File(
                label="Download Translation (Mono)", visible=False
            )
            output_file_dual = gr.File(
                label="Download Translation (Dual)", visible=False
            )

            translate_btn = gr.Button("Translate", variant="primary")
            cancel_btn = gr.Button("Cancel", variant="secondary")

            tech_details = gr.Markdown(
                tech_details_string,
                elem_classes=["secondary-text"],
            )

        with gr.Column(scale=2):
            gr.Markdown("## Preview")
            preview = PDF(label="Document Preview", visible=True, height=2000)

    # Event handlers
    def on_select_filetype(file_type):
        """Update visibility based on selected file type"""
        return (
            gr.update(visible=file_type == "File"),
            gr.update(visible=file_type == "Link"),
        )

    def on_select_page(choice):
        """Update page input visibility based on selection"""
        return gr.update(visible=choice == "Range")

    def on_select_service(service_name):
        """Update service-specific settings visibility"""
        if not detail_text_inputs:
            return
        detail_group_index = detail_text_input_index_map.get(service_name, [])
        if len(detail_text_inputs) == 1:
            return gr.update(visible=(0 in detail_group_index))
        else:
            return [
                gr.update(visible=(i in detail_group_index))
                for i in range(len(detail_text_inputs))
            ]

    def on_enhance_compatibility_change(enhance_value):
        """Update skip_clean and disable_rich_text_translate when enhance_compatibility changes"""
        if enhance_value:
            # When enhanced compatibility is enabled, both options are auto-enabled and disabled for user modification
            return (
                gr.update(value=True, interactive=False),
                gr.update(value=True, interactive=False),
            )
        else:
            # When disabled, allow user to modify these settings
            return (
                gr.update(interactive=True),
                gr.update(interactive=True),
            )

    def on_split_short_lines_change(split_value):
        """Update short_line_split_factor visibility based on split_short_lines value"""
        return gr.update(visible=split_value)

    # Default file handler
    file_input.upload(
        lambda x: x,
        inputs=file_input,
        outputs=preview,
    )

    # Event bindings
    file_type.select(
        on_select_filetype,
        file_type,
        [file_input, link_input],
    )

    page_range.select(
        on_select_page,
        page_range,
        page_input,
    )

    service.select(
        on_select_service,
        service,
        outputs=detail_text_inputs if len(detail_text_inputs) > 0 else None,
    )

    # Add event handler for enhance_compatibility
    enhance_compatibility.change(
        on_enhance_compatibility_change,
        enhance_compatibility,
        [skip_clean, disable_rich_text_translate],
    )

    # Add event handler for split_short_lines
    split_short_lines.change(
        on_split_short_lines_change,
        split_short_lines,
        short_line_split_factor,
    )

    # State for managing translation tasks
    state = gr.State({"session_id": None, "current_task": None})

    # Translation button click handler
    translate_btn.click(
        translate_file,
        inputs=[
            file_type,
            file_input,
            link_input,
            service,
            lang_from,
            lang_to,
            page_range,
            page_input,
            # PDF Output Options
            no_mono,
            no_dual,
            dual_translate_first,
            use_alternating_pages_dual,
            watermark_output_mode,
            # Advanced Options
            prompt,
            threads,
            min_text_length,
            rpc_doclayout,
            custom_system_prompt_input,
            skip_clean,
            disable_rich_text_translate,
            enhance_compatibility,
            split_short_lines,
            short_line_split_factor,
            translate_table_text,
            skip_scanned_detection,
            max_pages_per_part,
            formular_font_pattern,
            formular_char_pattern,
            ignore_cache,
            state,
            ocr_workaround,
            *translation_engine_arg_inputs,
        ],
        outputs=[
            output_file_mono,  # Mono PDF file
            preview,  # Preview
            output_file_dual,  # Dual PDF file
            output_file_mono,  # Visibility of mono output
            output_file_dual,  # Visibility of dual output
            output_title,  # Visibility of output title
        ],
    )

    # Cancel button click handler
    cancel_btn.click(
        stop_translate_file,
        inputs=[state],
    )


def parse_user_passwd(file_path: list) -> tuple[list, str]:
    """
    This function parses a user password file.

    Inputs:
        - file_path: The path to the file

    Returns:
        - A tuple containing the user list and HTML
    """
    content = ""
    if len(file_path) == 2:
        try:
            path = Path(file_path[1])
            content = path.read_text(encoding="utf-8")
        except FileNotFoundError:
            print(f"Error: File '{file_path[1]}' not found.")
    try:
        path = Path(file_path[0])
        tuple_list = [
            tuple(line.strip().split(","))
            for line in path.read_text(encoding="utf-8").splitlines()
            if line.strip()
        ]
    except FileNotFoundError:
        tuple_list = []
    return tuple_list, content


def setup_gui(
    share: bool = False,
    auth_file: list | None = None,
    server_port=7860,
    inbrowser: bool = True,
) -> None:
    """
    This function sets up the GUI for the application.

    Inputs:
        - share: Whether to share the GUI
        - auth_file: The authentication file
        - server_port: The port to run the server on

    Returns:
        - None
    """

    user_list = None
    html = None

    if auth_file:
        user_list, html = parse_user_passwd(auth_file)

    if not auth_file or not user_list:
        try:
            demo.launch(
                server_name="0.0.0.0",
                debug=True,
                inbrowser=inbrowser,
                share=share,
                server_port=server_port,
            )
        except Exception:
            print(
                "Error launching GUI using 0.0.0.0.\nThis may be caused by global mode of proxy software."
            )
            try:
                demo.launch(
                    server_name="127.0.0.1",
                    debug=True,
                    inbrowser=inbrowser,
                    share=share,
                    server_port=server_port,
                )
            except Exception:
                print(
                    "Error launching GUI using 127.0.0.1.\nThis may be caused by global mode of proxy software."
                )
                demo.launch(
                    debug=True, inbrowser=inbrowser, share=True, server_port=server_port
                )
    else:
        try:
            demo.launch(
                server_name="0.0.0.0",
                debug=True,
                inbrowser=inbrowser,
                share=share,
                auth=user_list,
                auth_message=html,
                server_port=server_port,
            )
        except Exception:
            print(
                "Error launching GUI using 0.0.0.0.\nThis may be caused by global mode of proxy software."
            )
            try:
                demo.launch(
                    server_name="127.0.0.1",
                    debug=True,
                    inbrowser=inbrowser,
                    share=share,
                    auth=user_list,
                    auth_message=html,
                    server_port=server_port,
                )
            except Exception:
                print(
                    "Error launching GUI using 127.0.0.1.\nThis may be caused by global mode of proxy software."
                )
                demo.launch(
                    debug=True,
                    inbrowser=inbrowser,
                    share=True,
                    auth=user_list,
                    auth_message=html,
                    server_port=server_port,
                )


# For auto-reloading while developing
if __name__ == "__main__":
    from rich.logging import RichHandler

    # disable httpx, openai, httpcore, http11 logs
    logging.getLogger("httpx").setLevel("CRITICAL")
    logging.getLogger("httpx").propagate = False
    logging.getLogger("openai").setLevel("CRITICAL")
    logging.getLogger("openai").propagate = False
    logging.getLogger("httpcore").setLevel("CRITICAL")
    logging.getLogger("httpcore").propagate = False
    logging.getLogger("http11").setLevel("CRITICAL")
    logging.getLogger("http11").propagate = False
    logging.basicConfig(level=logging.INFO, handlers=[RichHandler()])
    setup_gui(inbrowser=False)



import html
import logging
import re
import unicodedata

import requests
from pdf2zh_next.config.model import SettingsModel
from pdf2zh_next.translator.base_rate_limiter import BaseRateLimiter
from pdf2zh_next.translator.base_translator import BaseTranslator
from tenacity import before_sleep_log
from tenacity import retry
from tenacity import retry_if_exception_type
from tenacity import stop_after_attempt
from tenacity import wait_exponential

logger = logging.getLogger(__name__)


def remove_control_characters(s):
    return "".join(ch for ch in s if unicodedata.category(ch)[0] != "C")


class GoogleTranslator(BaseTranslator):
    name = "google"
    lang_map = {"zh": "zh-CN"}

    def __init__(
        self,
        settings: SettingsModel,
        rate_limiter: BaseRateLimiter,
    ):
        super().__init__(settings, rate_limiter)
        self.session = requests.Session()
        self.endpoint = "https://translate.google.com/m"
        self.headers = {
            "User-Agent": "Mozilla/4.0 (compatible;MSIE 6.0;Windows NT 5.1;SV1;.NET CLR 1.1.4322;.NET CLR 2.0.50727;.NET CLR 3.0.04506.30)"  # noqa: E501
        }

    @retry(
        retry=retry_if_exception_type(Exception),
        stop=stop_after_attempt(5),
        wait=wait_exponential(multiplier=1, min=1, max=15),
        before_sleep=before_sleep_log(logger, logging.WARNING),
    )
    def do_translate(self, text, rate_limit_params: dict = None):
        text = text[:5000]  # google translate max length
        response = self.session.get(
            self.endpoint,
            params={"tl": self.lang_out, "sl": self.lang_in, "q": text},
            headers=self.headers,
        )
        re_result = re.findall(
            r'(?s)class="(?:t0|result-container)">(.*?)<', response.text
        )
        if response.status_code == 400:
            result = "IRREPARABLE TRANSLATION ERROR"
        else:
            response.raise_for_status()
            result = html.unescape(re_result[0])
        return remove_control_characters(result)



from __future__ import annotations

from argparse import Namespace
from pathlib import Path
from unittest.mock import mock_open
from unittest.mock import patch

import pytest
import tomlkit
from pdf2zh_next.config.main import ConfigManager
from pdf2zh_next.config.main import MagicDefault
from pdf2zh_next.config.main import build_args_parser
from pydantic import BaseModel
from pydantic import Field


class NestedModel(BaseModel):
    field1: str
    field2: int


class DuplicateModel(BaseModel):
    field1: str
    nested: NestedModel = Field(default_factory=NestedModel)


class TestBuildArgsParser:
    def test_basic_parser_creation(self):
        """Test basic parser creation"""
        parser, field_name2type = build_args_parser()
        assert parser is not None
        assert field_name2type is not None
        assert isinstance(field_name2type, dict)

    def test_duplicate_field_detection(self):
        """Test duplicate field detection"""
        with pytest.raises(ValueError, match="duplicate field name"):
            build_args_parser(settings_model=DuplicateModel)

    def test_field_types(self):
        """Test different field types are handled correctly"""
        parser, field_name2type = build_args_parser()

        # Check some known fields
        assert "config_file" in field_name2type
        assert "basic" in field_name2type
        assert "translation" in field_name2type


class TestConfigManager:
    def test_singleton(self):
        """Test ConfigManager singleton pattern"""
        cm1 = ConfigManager()
        cm2 = ConfigManager()
        assert cm1 is cm2

    def test_parse_env_vars(self, monkeypatch: pytest.MonkeyPatch):
        """Test environment variable parsing"""
        # Set up test environment variables
        test_env = {
            "PDF2ZH_DEBUG": "true",
            "PDF2ZH_REPORT_INTERVAL": "0.5",
            "PDF2ZH_QPS": "10",
            "PDF2ZH_INPUT_FILES": '["test1.pdf", "test2.pdf"]',
            "PDF2ZH_INVALID_KEY": "value",  # Should be ignored
        }
        for key, value in test_env.items():
            monkeypatch.setenv(key, value)

        cm = ConfigManager()
        env_settings = cm.parse_env_vars()

        assert env_settings["basic"]["debug"] is True
        assert env_settings["report_interval"] == 0.5
        assert env_settings["translation"]["qps"] == 10
        assert "invalid_key" not in env_settings

    def test_convert_env_value(self):
        """Test environment variable value conversion"""
        cm = ConfigManager()

        # Test boolean conversion
        assert cm._convert_env_value("true", bool, None, ()) is True
        assert cm._convert_env_value("1", bool, None, ()) is True
        assert cm._convert_env_value("false", bool, None, ()) is False
        assert cm._convert_env_value("0", bool, None, ()) is False

        # Test numeric conversion
        assert cm._convert_env_value("42", int, None, ()) == 42
        assert cm._convert_env_value("3.14", float, None, ()) == 3.14

        # Test string passthrough
        assert cm._convert_env_value("test", str, None, ()) == "test"

        # Test invalid conversions
        with pytest.raises(ValueError):
            cm._convert_env_value("not_a_number", int, None, ())

    def test_deep_merge(self):
        """Test dictionary deep merge"""
        cm = ConfigManager()

        # Test basic merge
        target = {"a": 1, "b": 2}
        source = {"b": 3, "c": 4}
        result = cm._deep_merge(target, source)
        assert result == {"a": 1, "b": 3, "c": 4}

        # Test nested merge
        target = {"a": {"x": 1, "y": 2}, "b": 3}
        source = {"a": {"y": 3, "z": 4}, "c": 5}
        result = cm._deep_merge(target, source)
        assert result == {"a": {"x": 1, "y": 3, "z": 4}, "b": 3, "c": 5}

        # Test list merge
        target = {"a": [1, 2], "b": 3}
        source = {"a": [3, 4], "c": 5}
        result = cm._deep_merge(target, source)
        assert result == {"a": [1, 2, 3, 4], "b": 3, "c": 5}

    def test_merge_settings(self):
        """Test settings merge with priority"""
        cm = ConfigManager()

        cli_settings = {"debug": True, "qps": 10}
        env_settings = {"debug": False, "report_interval": 0.5}
        default_settings = {"debug": False, "qps": 5, "report_interval": 0.1}

        result = cm.merge_settings([cli_settings, env_settings, default_settings])

        # CLI settings should take precedence
        assert result["debug"] is True
        assert result["qps"] == 10
        # Env settings should take precedence over defaults
        assert result["report_interval"] == 0.5

    def test_initialize_config(self, monkeypatch: pytest.MonkeyPatch):
        """Test complete configuration initialization"""
        # Set up environment variables
        monkeypatch.setenv("PDF2ZH_REPORT_INTERVAL", "0.5")
        monkeypatch.setenv("PDF2ZH_QPS", "10")

        # Create a ConfigManager instance
        cm = ConfigManager()

        # Mock command line arguments
        args = Namespace(
            debug=True,
            report_interval=MagicDefault,  # Should use env var
            qps=15,  # Should override env var
            google=True,
        )

        # Initialize configuration
        with patch("argparse.ArgumentParser.parse_args", return_value=args):
            config = cm.initialize_config()

        # Verify settings
        assert cm.settings.basic.debug is True
        assert cm.settings.translation.qps == 15  # CLI value should override env var
        assert cm.settings.report_interval == 0.5  # Should use env var
        assert cm.settings is not None

    @pytest.fixture
    def sample_toml_content(self) -> str:
        """Sample TOML configuration content"""
        return """
        debug = true
        report_interval = 0.5
        qps = 10
        """

    @pytest.fixture
    def temp_config_dir(self, tmp_path: Path) -> Path:
        """Create a temporary config directory"""
        config_dir = tmp_path / ".config" / "pdf2zh_next"
        config_dir.mkdir(parents=True)
        return config_dir

    def test_read_toml_file(self, sample_toml_content: str):
        """Test reading TOML configuration file"""
        cm = ConfigManager()

        # Test successful read
        with patch("pathlib.Path.open", mock_open(read_data=sample_toml_content)):
            config = cm._read_toml_file(Path("test.toml"))
            assert config["debug"] is True
            assert config["report_interval"] == 0.5
            assert config["qps"] == 10

        # Test file not found
        with patch("pathlib.Path.open", side_effect=FileNotFoundError):
            config = cm._read_toml_file(Path("nonexistent.toml"))
            assert config == {}

        # Test invalid TOML
        with patch("pathlib.Path.open", mock_open(read_data="invalid = toml [")):
            config = cm._read_toml_file(Path("invalid.toml"))
            assert config == {}

    def test_write_toml_file(self, temp_config_dir: Path):
        """Test writing TOML configuration file"""
        cm = ConfigManager()
        test_file = temp_config_dir / "test.toml"
        test_content = {"debug": True, "report_interval": 0.5}

        # Test successful write
        cm._write_toml_file(test_file, test_content)
        assert test_file.exists()

        # Verify content
        with test_file.open() as f:
            loaded_content = tomlkit.load(f)
            assert dict(loaded_content) == test_content

    def test_is_file_content_identical(self, sample_toml_content: str):
        """Test comparing file content with given content"""
        cm = ConfigManager()
        test_content = {"debug": True, "report_interval": 0.5, "qps": 10}

        # Test identical content
        with patch("pathlib.Path.open", mock_open(read_data=sample_toml_content)):
            assert cm._is_file_content_identical(Path("test.toml"), test_content)

        # Test different content
        different_content = {"debug": False, "report_interval": 1.0}
        assert not cm._is_file_content_identical(Path("test.toml"), different_content)

    def test_update_version_default_config(self, temp_config_dir: Path):
        """Test updating version default configuration file"""
        cm = ConfigManager()
        version_file = temp_config_dir / "version.default.toml"

        with (
            patch.object(ConfigManager, "_ensure_config_dir"),
            patch("pdf2zh_next.config.main.VERSION_DEFAULT_CONFIG_FILE", version_file),
        ):
            # First update should create the file
            cm._update_version_default_config()
            assert version_file.exists()

            # Test that calling update again doesn't raise errors
            # We can't reliably check for identical mtime due to platform differences
            # and the speed of test execution, so just verify it executes without errors
            cm._update_version_default_config()
            assert version_file.exists()  # File should still exist

    def test_config_file_priority(
        self, monkeypatch: pytest.MonkeyPatch, temp_config_dir: Path
    ):
        """Test configuration file priority order"""
        cm = ConfigManager()

        # Create test files
        user_config = temp_config_dir / "user.toml"
        default_config = temp_config_dir / "default.toml"

        # Set up test configurations
        user_config_content = {"basic": {"debug": True}, "translation": {"qps": 20}}
        default_config_content = {
            "basic": {"debug": True},
            "translation": {"qps": 20},
            "report_interval": 0.5,
        }

        with patch("pdf2zh_next.const.DEFAULT_CONFIG_FILE", default_config):
            # Write test configurations
            cm._write_toml_file(user_config, user_config_content)
            cm._write_toml_file(default_config, default_config_content)

            # Set up environment and CLI args
            monkeypatch.setenv("PDF2ZH_REPORT_INTERVAL", "1.0")
            args = Namespace(
                config_file=str(user_config),
                debug=MagicDefault,
                qps=MagicDefault,
                report_interval=MagicDefault,
            )
            cm.default_config_file_path = default_config
            # Initialize configuration
            with patch("argparse.ArgumentParser.parse_args", return_value=args):
                cm.initialize_config()

            # Verify priority: CLI > Env > User Config > Default Config
            assert cm.settings.basic.debug is True  # from user config
            assert cm.settings.translation.qps == 20  # from user config
            assert cm.settings.report_interval == 1.0  # from env var

    def test_ensure_config_dir(self, temp_config_dir: Path):
        """Test configuration directory creation"""
        cm = ConfigManager()
        test_dir = temp_config_dir / "subdir"

        with patch("pdf2zh_next.config.main.DEFAULT_CONFIG_DIR", test_dir):
            # Directory should not exist initially
            assert not test_dir.exists()

            # Create directory
            cm._ensure_config_dir()
            assert test_dir.exists()
            assert test_dir.is_dir()

            # Should not raise error when directory already exists
            cm._ensure_config_dir()
            assert test_dir.exists()

    def test_process_toml_content(self):
        """Test processing of TOML content with various data types and structures"""
        cm = ConfigManager()

        # Test basic types
        content = {
            "string": "value",
            "number": 42,
            "float": 3.14,
            "bool": True,
            "null_str": "null",
        }
        processed = cm._process_toml_content(content)
        assert processed["string"] == "value"
        assert processed["number"] == 42
        assert processed["float"] == 3.14
        assert processed["bool"] is True
        assert processed["null_str"] is None

        # Test nested structures
        nested_content = {
            "level1": {
                "level2": {"value": "null", "list": [1, 2, 3], "nested_null": "null"}
            }
        }
        processed = cm._process_toml_content(nested_content)
        assert processed["level1"]["level2"]["value"] is None
        assert processed["level1"]["level2"]["list"] == [1, 2, 3]
        assert processed["level1"]["level2"]["nested_null"] is None

    def test_complex_env_vars(self, monkeypatch: pytest.MonkeyPatch):
        """Test parsing of complex environment variables"""
        cm = ConfigManager()

        # Test basic types first
        monkeypatch.setenv("PDF2ZH_DEBUG", "true")
        monkeypatch.setenv("PDF2ZH_QPS", "20")
        env_settings = cm.parse_env_vars()
        assert "basic" in env_settings and "debug" in env_settings["basic"]
        assert env_settings["basic"]["debug"] is True
        assert env_settings["translation"]["qps"] == 20

        # Test optional types
        monkeypatch.setenv("PDF2ZH_REPORT_INTERVAL", "0.5")
        env_settings = cm.parse_env_vars()
        assert env_settings["report_interval"] == 0.5

    def test_get_default_config(self):
        """Test default configuration generation"""
        cm = ConfigManager()
        default_config = cm._get_default_config()

        # Verify basic structure
        assert isinstance(default_config, dict)

        # Verify input_files is converted to list
        input_files_key = [
            k for k in default_config.keys() if k.endswith("input_files")
        ]
        if input_files_key:
            assert isinstance(default_config[input_files_key[0]], list)

        # Verify essential configuration fields
        assert "basic" in default_config and "debug" in default_config["basic"]
        assert (
            "translation" in default_config and "qps" in default_config["translation"]
        )
        assert isinstance(default_config["basic"]["debug"], bool)
        assert isinstance(default_config["translation"]["qps"], int | float)

    def test_settings_not_initialized(self):
        """Test accessing settings before initialization"""
        cm = ConfigManager()
        cm._settings = None  # Force settings to be None
        with pytest.raises(RuntimeError, match="Settings not initialized"):
            _ = cm.settings

    def test_initialize_config_with_invalid_file(self, monkeypatch: pytest.MonkeyPatch):
        """Test initialization with invalid config file"""
        cm = ConfigManager()

        # Create args with non-existent config file
        args = Namespace(
            config_file="nonexistent.toml",
            debug=MagicDefault,
            report_interval=MagicDefault,
        )

        # Mock command line arguments
        with patch("argparse.ArgumentParser.parse_args", return_value=args):
            # Should not raise exception, should use default values
            cm.initialize_config()
            assert cm.settings is not None

    def test_initialize_config_with_invalid_toml(self, temp_config_dir: Path):
        """Test initialization with invalid TOML content"""
        cm = ConfigManager()
        invalid_config = temp_config_dir / "invalid.toml"

        # Create invalid TOML file
        invalid_config.write_text("invalid = toml [ content")

        args = Namespace(
            config_file=str(invalid_config),
            debug=MagicDefault,
            report_interval=MagicDefault,
        )

        # Mock command line arguments
        with patch("argparse.ArgumentParser.parse_args", return_value=args):
            # Should not raise exception, should use default values
            cm.initialize_config()
            assert cm.settings is not None

    def test_nested_config_priority(self, monkeypatch: pytest.MonkeyPatch):
        """Test priority handling with nested configurations"""
        cm = ConfigManager()

        # Set up nested configurations at different levels
        cli_args = {"translation": {"model": "gpt-4", "temperature": 0.8}}

        env_vars = {
            "translation": {"model": "gpt-3.5", "temperature": 0.7, "max_tokens": 1000}
        }

        default_config = {
            "translation": {
                "model": "gpt-3",
                "temperature": 0.5,
                "max_tokens": 500,
                "timeout": 30,
            }
        }

        # Test merging with priority
        result = cm.merge_settings([cli_args, env_vars, default_config])

        # CLI values should take precedence
        assert result["translation"]["model"] == "gpt-4"
        assert result["translation"]["temperature"] == 0.8

        # Env vars should take precedence over defaults
        assert result["translation"]["max_tokens"] == 1000

        # Default values should be preserved if not overridden
        assert result["translation"]["timeout"] == 30

    def test_parse_cli_args(self):
        """Test command line argument parsing"""
        cm = ConfigManager()

        # Test with basic argument types
        test_args = [
            "--debug",
            "--qps",
            "20",
            "--report-interval",
            "0.5",
            "file1.pdf",  # Single file for now
        ]

        with patch("sys.argv", ["script.py"] + test_args):
            cm.parse_cli_args()

            assert cm.settings.basic.debug is True
            assert cm.settings.translation.qps == 20
            assert cm.settings.report_interval == 0.5
            # input_files should be empty set by default
            assert isinstance(cm.settings.basic.input_files, set)
            assert len(cm.settings.basic.input_files) == 1

        # Test with input files
        test_args_with_files = ["--debug", "file1.pdf", "file2.pdf"]

        with patch("sys.argv", ["script.py"] + test_args_with_files):
            cm.parse_cli_args()
            assert isinstance(cm.settings.basic.input_files, set)
            assert "file1.pdf" in cm.settings.basic.input_files
            assert "file2.pdf" in cm.settings.basic.input_files

    def test_end_to_end_config(
        self, temp_config_dir: Path, monkeypatch: pytest.MonkeyPatch
    ):
        """Test end-to-end configuration with all sources"""
        cm = ConfigManager()

        # Create test config file
        config_file = temp_config_dir / "test_config.toml"
        test_input_pdf = temp_config_dir / "test.pdf"
        with test_input_pdf.open("w") as f:
            f.write("1")
        config_content = """
        [basic]
        debug = false
        
        [translation]
        model = "gpt-3"
        qps = 5
        """
        config_file.write_text(config_content)

        # Set environment variables
        monkeypatch.setenv("PDF2ZH_QPS", "10")
        monkeypatch.setenv("PDF2ZH_REPORT_INTERVAL", "1.0")

        # Create CLI arguments
        args = Namespace(
            config_file=str(config_file),
            debug=True,  # Should override config file
            qps=MagicDefault,  # Should use env var
            report_interval=MagicDefault,  # Should use env var
            input_files={test_input_pdf.as_posix()},  # Add a test file
        )

        # Initialize with all sources
        with patch("argparse.ArgumentParser.parse_args", return_value=args):
            cm.initialize_config()

            # Verify priority order
            assert cm.settings.basic.debug is True  # From CLI
            assert cm.settings.translation.qps == 10  # From env var
            assert cm.settings.report_interval == 1.0  # From env var
            assert (
                test_input_pdf.as_posix() in cm.settings.basic.input_files
            )  # From CLI args

    def test_file_system_operations(self, temp_config_dir: Path):
        """Test actual file system operations"""
        cm = ConfigManager()

        # Test directory creation
        test_dir = temp_config_dir / "new_config_dir"
        with patch("pdf2zh_next.config.main.DEFAULT_CONFIG_DIR", test_dir):
            cm._ensure_config_dir()
            assert test_dir.exists()
            assert test_dir.is_dir()

        # Test file writing and reading
        test_file = test_dir / "test.toml"
        test_content = {"basic": {"debug": True}, "translation": {"qps": 15}}

        # Write file
        cm._write_toml_file(test_file, test_content)
        assert test_file.exists()

        # Read and verify content
        read_content = cm._read_toml_file(test_file)
        assert read_content["basic"]["debug"] is True
        assert read_content["translation"]["qps"] == 15



__version__ = "2.0.0.rc1"
__major_version__ = "2"
__config_file_version__ = "3"

from pathlib import Path

# Constants for configuration paths
DEFAULT_CONFIG_DIR = Path("~/.config/pdf2zh").expanduser()
DEFAULT_CONFIG_DIR.mkdir(parents=True, exist_ok=True)
DEFAULT_CONFIG_FILE = DEFAULT_CONFIG_DIR / f"config.v{__config_file_version__}.toml"
WRITE_TEMP_CONFIG_FILE = (
    DEFAULT_CONFIG_DIR / f"config.v{__config_file_version__}.temp.toml"
)
VERSION_DEFAULT_CONFIG_DIR = DEFAULT_CONFIG_DIR / "default"
VERSION_DEFAULT_CONFIG_DIR.mkdir(parents=True, exist_ok=True)
VERSION_DEFAULT_CONFIG_FILE = VERSION_DEFAULT_CONFIG_DIR / f"{__version__}.toml"



import logging

import openai
from babeldoc.document_il.utils.atomic_integer import AtomicInteger
from pdf2zh_next.config.model import SettingsModel
from pdf2zh_next.translator.base_rate_limiter import BaseRateLimiter
from pdf2zh_next.translator.base_translator import BaseTranslator
from tenacity import before_sleep_log
from tenacity import retry
from tenacity import retry_if_exception_type
from tenacity import stop_after_attempt
from tenacity import wait_exponential

logger = logging.getLogger(__name__)


class AzureOpenAITranslator(BaseTranslator):
    name = "azure-openai"

    def __init__(
        self,
        settings: SettingsModel,
        rate_limiter: BaseRateLimiter,
    ):
        super().__init__(settings, rate_limiter)
        self.options = {"temperature": 0}  # 随机采样可能会打断公式标记
        self.client = openai.AzureOpenAI(
            azure_endpoint=settings.translate_engine_settings.azure_openai_base_url,
            azure_deployment=settings.translate_engine_settings.azure_openai_model,
            api_version=settings.translate_engine_settings.azure_openai_api_version,
            api_key=settings.translate_engine_settings.azure_openai_api_key,
        )
        self.add_cache_impact_parameters("temperature", self.options["temperature"])
        self.model = settings.translate_engine_settings.azure_openai_model
        self.add_cache_impact_parameters("model", self.model)
        self.add_cache_impact_parameters("prompt", self.prompt(""))
        self.token_count = AtomicInteger()
        self.prompt_token_count = AtomicInteger()
        self.completion_token_count = AtomicInteger()

    @retry(
        retry=retry_if_exception_type(openai.RateLimitError),
        stop=stop_after_attempt(100),
        wait=wait_exponential(multiplier=1, min=1, max=15),
        before_sleep=before_sleep_log(logger, logging.WARNING),
    )
    def do_translate(self, text, rate_limit_params: dict = None) -> str:
        response = self.client.chat.completions.create(
            model=self.model,
            **self.options,
            messages=self.prompt(text),
        )
        self.token_count.inc(response.usage.total_tokens)
        self.prompt_token_count.inc(response.usage.prompt_tokens)
        self.completion_token_count.inc(response.usage.completion_tokens)
        message = response.choices[0].message.content.strip()
        message = self._remove_cot_content(message)
        return message

    @retry(
        retry=retry_if_exception_type(openai.RateLimitError),
        stop=stop_after_attempt(100),
        wait=wait_exponential(multiplier=1, min=1, max=15),
        before_sleep=before_sleep_log(logger, logging.WARNING),
    )
    def do_llm_translate(self, text, rate_limit_params: dict = None):
        if text is None:
            return None

        response = self.client.chat.completions.create(
            model=self.model,
            **self.options,
            messages=[
                {
                    "role": "user",
                    "content": text,
                },
            ],
        )
        self.token_count.inc(response.usage.total_tokens)
        self.prompt_token_count.inc(response.usage.prompt_tokens)
        self.completion_token_count.inc(response.usage.completion_tokens)
        message = response.choices[0].message.content.strip()
        message = self._remove_cot_content(message)
        return message



from pdf2zh_next.config import AnythingLLMSettings
from pdf2zh_next.config import AzureOpenAISettings
from pdf2zh_next.config import AzureSettings
from pdf2zh_next.config import BingSettings
from pdf2zh_next.config import DeepLSettings
from pdf2zh_next.config import DeepSeekSettings
from pdf2zh_next.config import DifySettings
from pdf2zh_next.config import GeminiSettings
from pdf2zh_next.config import GoogleSettings
from pdf2zh_next.config import GrokSettings
from pdf2zh_next.config import GroqSettings
from pdf2zh_next.config import ModelScopeSettings
from pdf2zh_next.config import OllamaSettings
from pdf2zh_next.config import OpenAISettings
from pdf2zh_next.config import QwenMtSettings
from pdf2zh_next.config import SiliconFlowSettings
from pdf2zh_next.config import TencentSettings
from pdf2zh_next.config import XinferenceSettings
from pdf2zh_next.config import ZhipuSettings
from pdf2zh_next.config.main import ConfigManager
from pdf2zh_next.config.model import BasicSettings
from pdf2zh_next.config.model import PDFSettings
from pdf2zh_next.config.model import SettingsModel
from pdf2zh_next.config.model import TranslationSettings
from pdf2zh_next.config.model import WatermarkOutputMode
from pdf2zh_next.high_level import create_babeldoc_config
from pdf2zh_next.high_level import do_translate_async_stream
from pdf2zh_next.high_level import do_translate_file
from pdf2zh_next.high_level import do_translate_file_async

# from pdf2zh_next.high_level import translate, translate_stream

__version__ = "2.0.0.rc1"
__author__ = "Byaidu, awwaawwa"
__license__ = "AGPL-3.0"
__maintainer__ = "awwaawwa"
__email__ = "aw@funstory.ai"

__all__ = [
    "SettingsModel",
    "BasicSettings",
    "OpenAISettings",
    "BingSettings",
    "GoogleSettings",
    "DeepLSettings",
    "DeepSeekSettings",
    "OllamaSettings",
    "XinferenceSettings",
    "AzureOpenAISettings",
    "ModelScopeSettings",
    "ZhipuSettings",
    "SiliconFlowSettings",
    "TencentSettings",
    "GeminiSettings",
    "AzureSettings",
    "AnythingLLMSettings",
    "DifySettings",
    "GrokSettings",
    "GroqSettings",
    "QwenMtSettings",
    "PDFSettings",
    "TranslationSettings",
    "WatermarkOutputMode",
    "do_translate_file_async",
    "do_translate_file",
    "do_translate_async_stream",
    "create_babeldoc_config",
    "ConfigManager",
]



import importlib
import logging

from pdf2zh_next.config.model import SettingsModel
from pdf2zh_next.config.translate_engine_model import (
    NOT_SUPPORTED_TRANSLATION_ENGINE_SETTING_TYPE,
)
from pdf2zh_next.config.translate_engine_model import TRANSLATION_ENGINE_METADATA
from pdf2zh_next.config.translate_engine_model import TranslateEngineSettingError
from pdf2zh_next.translator.base_rate_limiter import BaseRateLimiter
from pdf2zh_next.translator.base_translator import BaseTranslator
from pdf2zh_next.translator.rate_limiter.qps_rate_limiter import QPSRateLimiter

logger = logging.getLogger(__name__)


def get_rate_limiter(settings: SettingsModel) -> BaseRateLimiter:
    if settings.translation.qps:
        return QPSRateLimiter(settings.translation.qps)
    else:
        return None


def get_translator(settings: SettingsModel) -> BaseTranslator:
    rate_limiter = get_rate_limiter(settings=settings)
    translator_config = settings.translate_engine_settings

    if isinstance(translator_config, NOT_SUPPORTED_TRANSLATION_ENGINE_SETTING_TYPE):
        raise TranslateEngineSettingError(
            f"{translator_config.translate_engine_type} is not supported, Please use other translator!"
        )

    for metadata in TRANSLATION_ENGINE_METADATA:
        if isinstance(translator_config, metadata.setting_model_type):
            translate_engine_type = metadata.translate_engine_type
            logger.info(f"Using {translate_engine_type} translator")
            model_name = f"pdf2zh_next.translator.translator_impl.{translate_engine_type.lower()}"
            module = importlib.import_module(model_name)
            return getattr(module, f"{translate_engine_type}Translator")(
                settings, rate_limiter
            )

    raise ValueError("No translator found")



from pdf2zh_next.config.main import ConfigManager
from pdf2zh_next.config.model import BasicSettings
from pdf2zh_next.config.model import PDFSettings
from pdf2zh_next.config.model import SettingsModel
from pdf2zh_next.config.model import TranslationSettings
from pdf2zh_next.config.model import WatermarkOutputMode
from pdf2zh_next.config.translate_engine_model import TRANSLATION_ENGINE_METADATA
from pdf2zh_next.config.translate_engine_model import AnythingLLMSettings
from pdf2zh_next.config.translate_engine_model import AzureOpenAISettings
from pdf2zh_next.config.translate_engine_model import AzureSettings
from pdf2zh_next.config.translate_engine_model import BingSettings
from pdf2zh_next.config.translate_engine_model import DeepLSettings
from pdf2zh_next.config.translate_engine_model import DeepSeekSettings
from pdf2zh_next.config.translate_engine_model import DifySettings
from pdf2zh_next.config.translate_engine_model import GeminiSettings
from pdf2zh_next.config.translate_engine_model import GoogleSettings
from pdf2zh_next.config.translate_engine_model import GrokSettings
from pdf2zh_next.config.translate_engine_model import GroqSettings
from pdf2zh_next.config.translate_engine_model import ModelScopeSettings
from pdf2zh_next.config.translate_engine_model import OllamaSettings
from pdf2zh_next.config.translate_engine_model import OpenAISettings
from pdf2zh_next.config.translate_engine_model import QwenMtSettings
from pdf2zh_next.config.translate_engine_model import SiliconFlowSettings
from pdf2zh_next.config.translate_engine_model import TencentSettings
from pdf2zh_next.config.translate_engine_model import XinferenceSettings
from pdf2zh_next.config.translate_engine_model import ZhipuSettings

__all__ = [
    "ConfigManager",
    "SettingsModel",
    "BasicSettings",
    "TranslationSettings",
    "PDFSettings",
    "WatermarkOutputMode",
    "BingSettings",
    "GoogleSettings",
    "OpenAISettings",
    "DeepLSettings",
    "OllamaSettings",
    "XinferenceSettings",
    "AzureOpenAISettings",
    "ModelScopeSettings",
    "ZhipuSettings",
    "SiliconFlowSettings",
    "TencentSettings",
    "GeminiSettings",
    "AzureSettings",
    "AnythingLLMSettings",
    "DifySettings",
    "GrokSettings",
    "GroqSettings",
    "QwenMtSettings",
    "DeepSeekSettings",
    "TRANSLATION_ENGINE_METADATA",
]



import logging

from azure.ai.translation.text import TextTranslationClient
from azure.core.credentials import AzureKeyCredential
from pdf2zh_next.config.model import SettingsModel
from pdf2zh_next.translator.base_rate_limiter import BaseRateLimiter
from pdf2zh_next.translator.base_translator import BaseTranslator
from tenacity import before_sleep_log
from tenacity import retry
from tenacity import retry_if_exception_type
from tenacity import stop_after_attempt
from tenacity import wait_exponential

logger = logging.getLogger(__name__)


class AzureTranslator(BaseTranslator):
    name = "azure"
    lang_map = {"zh": "zh-Hans"}

    def __init__(
        self,
        settings: SettingsModel,
        rate_limiter: BaseRateLimiter,
    ):
        super().__init__(settings, rate_limiter)
        endpoint = settings.translate_engine_settings.azure_endpoint
        api_key = settings.translate_engine_settings.azure_api_key
        credential = AzureKeyCredential(api_key)
        self.client = TextTranslationClient(
            endpoint=endpoint, credential=credential, region="chinaeast2"
        )

    @retry(
        retry=retry_if_exception_type(Exception),
        stop=stop_after_attempt(5),
        wait=wait_exponential(multiplier=1, min=1, max=15),
        before_sleep=before_sleep_log(logger, logging.WARNING),
    )
    def do_translate(self, text, rate_limit_params: dict = None):
        response = self.client.translate(
            body=[text],
            from_language=self.lang_in,
            to_language=[self.lang_out],
        )
        translated_text = response[0].translations[0].text
        return translated_text



import json
import logging
from pathlib import Path

from peewee import SQL
from peewee import AutoField
from peewee import CharField
from peewee import Model
from peewee import SqliteDatabase
from peewee import TextField

# we don't init the database here
db = SqliteDatabase(None)
logger = logging.getLogger(__name__)


class _TranslationCache(Model):
    id = AutoField()
    translate_engine = CharField(max_length=20)
    translate_engine_params = TextField()
    original_text = TextField()
    translation = TextField()

    class Meta:
        database = db
        constraints = [
            SQL(
                """
            UNIQUE (
                translate_engine,
                translate_engine_params,
                original_text
                )
            ON CONFLICT REPLACE
            """
            )
        ]


class TranslationCache:
    @staticmethod
    def _sort_dict_recursively(obj):
        if isinstance(obj, dict):
            return {
                k: TranslationCache._sort_dict_recursively(v)
                for k in sorted(obj.keys())
                for v in [obj[k]]
            }
        elif isinstance(obj, list):
            return [TranslationCache._sort_dict_recursively(item) for item in obj]
        return obj

    def __init__(self, translate_engine: str, translate_engine_params: dict = None):
        assert len(translate_engine) < 20, (
            "current cache require translate engine name less than 20 characters"
        )
        self.translate_engine = translate_engine
        self.replace_params(translate_engine_params)

    # The program typically starts multi-threaded translation
    # only after cache parameters are fully configured,
    # so thread safety doesn't need to be considered here.
    def replace_params(self, params: dict = None):
        if params is None:
            params = {}
        self.params = params
        params = self._sort_dict_recursively(params)
        self.translate_engine_params = json.dumps(params)

    def update_params(self, params: dict = None):
        if params is None:
            params = {}
        self.params.update(params)
        self.replace_params(self.params)

    def add_params(self, k: str, v):
        self.params[k] = v
        self.replace_params(self.params)

    # Since peewee and the underlying sqlite are thread-safe,
    # get and set operations don't need locks.
    def get(self, original_text: str) -> str | None:
        result = _TranslationCache.get_or_none(
            translate_engine=self.translate_engine,
            translate_engine_params=self.translate_engine_params,
            original_text=original_text,
        )
        return result.translation if result else None

    def set(self, original_text: str, translation: str):
        try:
            _TranslationCache.create(
                translate_engine=self.translate_engine,
                translate_engine_params=self.translate_engine_params,
                original_text=original_text,
                translation=translation,
            )
        except Exception as e:
            logger.debug(f"Error setting cache: {e}")


def init_db(remove_exists=False):
    cache_folder = Path.home() / ".cache" / "pdf2zh_next"
    cache_folder.mkdir(parents=True, exist_ok=True)
    # The current version does not support database migration, so add the version number to the file name.
    cache_db_path = cache_folder / "cache.v1.db"
    if remove_exists and cache_db_path.exists():
        cache_db_path.unlink()
    db.init(
        str(cache_db_path),
        pragmas={
            "journal_mode": "wal",
            "busy_timeout": 1000,
        },
    )
    db.create_tables([_TranslationCache], safe=True)


def init_test_db():
    import os
    import tempfile

    fd, cache_db_path = tempfile.mkstemp(suffix=".db")
    os.close(fd)  # Close the file descriptor as we only need the path
    test_db = SqliteDatabase(
        cache_db_path,
        pragmas={
            "journal_mode": "wal",
            "busy_timeout": 1000,
        },
    )
    test_db.bind([_TranslationCache], bind_refs=False, bind_backrefs=False)
    test_db.connect()
    test_db.create_tables([_TranslationCache], safe=True)
    return test_db


def clean_test_db(test_db):
    test_db.drop_tables([_TranslationCache])
    test_db.close()
    db_path = Path(test_db.database)
    if db_path.exists():
        db_path.unlink()
    wal_path = Path(str(db_path) + "-wal")
    if wal_path.exists():
        wal_path.unlink()
    shm_path = Path(str(db_path) + "-shm")
    if shm_path.exists():
        shm_path.unlink()


init_db()



from __future__ import annotations

import logging

from pydantic import Field
from pydantic import create_model

from pdf2zh_next.config.model import SettingsModel
from pdf2zh_next.config.translate_engine_model import _DEFAULT_TRANSLATION_ENGINE
from pdf2zh_next.config.translate_engine_model import TRANSLATION_ENGINE_METADATA

logger = logging.getLogger(__name__)

# The following is magic code,
# if you need to modify it,
# please contact the maintainer!

__translation_flag_fields = {
    x.cli_flag_name: (
        bool,
        Field(
            default=False, description=f"Use {x.translate_engine_type} for translation"
        ),
    )
    for x in TRANSLATION_ENGINE_METADATA
}

__translation_flag_fields.update(
    {
        x.cli_detail_field_name: (
            x.setting_model_type,
            Field(default_factory=x.setting_model_type),
        )
        for x in TRANSLATION_ENGINE_METADATA
        if x.cli_detail_field_name
    }
)

__exclude_fields = list(__translation_flag_fields.keys())

# If you want to use more field parameters in `pdf2zh_next/config/model.py`
# please add the corresponding forwarding here!

__cli_env_settings_model_fields = {
    k: (
        v.annotation,
        Field(
            default=v.default,
            description=v.description,
            default_factory=v.default_factory,
            alias=v.alias,
            discriminator=v.discriminator,
        ),
    )
    for k, v in SettingsModel.model_fields.items()
    if k != "translate_engine_settings"
}
__cli_env_settings_model_fields.update(__translation_flag_fields)

CLIEnvSettingsModel = create_model(
    "CLIEnvSettingsModel",
    **__cli_env_settings_model_fields,
)


def to_settings_model(self) -> SettingsModel:
    for metadata in TRANSLATION_ENGINE_METADATA:
        if getattr(self, metadata.cli_flag_name):
            if metadata.cli_detail_field_name:
                translate_engine_settings = metadata.setting_model_type(
                    **getattr(self, metadata.cli_detail_field_name).model_dump()
                )
            else:
                translate_engine_settings = metadata.setting_model_type()
            break
    else:
        logger.warning("No translation engine selected, using Bing")
        translate_engine_settings = _DEFAULT_TRANSLATION_ENGINE()

    return SettingsModel(
        **self.model_dump(exclude=__exclude_fields),
        translate_engine_settings=translate_engine_settings,
    )


def validate_settings(self) -> None:
    self.to_settings_model().validate_settings()


def clone(self):
    return self.model_copy(deep=True)


CLIEnvSettingsModel.to_settings_model = to_settings_model
CLIEnvSettingsModel.validate_settings = validate_settings
CLIEnvSettingsModel.clone = clone



import threading
import time

from pdf2zh_next.translator.base_rate_limiter import BaseRateLimiter


class QPSRateLimiter(BaseRateLimiter):
    def __init__(self, max_qps: int):
        self.max_qps = max_qps
        self.min_interval = 1.0 / max_qps
        self.last_requests = []  # Track last N requests
        self.window_size = max_qps  # Track requests in a sliding window
        self.lock = threading.Lock()

    def wait(self, _rate_limit_params: dict = None):
        with self.lock:
            now = time.time()

            # Clean up old requests outside the 1-second window
            while self.last_requests and now - self.last_requests[0] > 1.0:
                self.last_requests.pop(0)

            # If we have less than max_qps requests in the last second, allow immediately
            if len(self.last_requests) < self.max_qps:
                self.last_requests.append(now)
                return

            # Otherwise, wait until we can make the next request
            next_time = self.last_requests[0] + 1.0
            if next_time > now:
                time.sleep(next_time - now)
            self.last_requests.pop(0)
            self.last_requests.append(next_time)

    def set_max_qps(self, max_qps):
        self.max_qps = max_qps
        self.min_interval = 1.0 / max_qps
        self.window_size = max_qps
